#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <cuda.h>
#include <sys/time.h>
#include <math.h>

#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}

__global__ void cuda_saxpb(const float *xd, const float a, const float b, 
                           float *yd, const int n) {

	int i = threadIdx.x + blockIdx.x*blockDim.x;
	if (i<n) {
		yd[i] = a*xd[i]+b;
	}
	return;
}

void cpu_saxpb(const float *x, float a, float b, float *y, int n) {

	int i;
	for (i=0;i<n;i++) {
		y[i] = a*x[i]+b;
	}
	return;
}

int get_options(int argc, char **argv, int *n, int *nb, float *a, float *b, int *niters);
void tick(struct timeval *timer);
double tock(struct timeval *timer);

int main(int argc, char **argv) {
	int n=1000;
	int nblocks=10;
	float a = 5.;
	float b = -1.;
	int err;
	float *x, *y, *ycuda;
	float *xd, *yd;
	double abserr;
	int blocksize;
    int i;
    int niters=1;
    struct timeval cputimer;
    struct timeval gputimer;
    double cputime, gputime;

	err = get_options(argc, argv, &n, &nblocks, &a, &b, &niters);
	if (err) return 0;

	x = (float *)malloc(n*sizeof(float));
	if (!x) return 1;
		
	y = (float *)malloc(n*sizeof(float));
	if (!y) {free(x); return 1;}

	ycuda = (float *)malloc(n*sizeof(float));
	if (!ycuda) {free(y); free(x); return 1;}


	/* run CPU code */

    tick(&cputimer);
    for (i=0; i<niters;i++)
        cpu_saxpb(x, a, b, y, n);
    cputime = tock(&cputimer);

	/* run GPU code */
	CHK_CUDA( cudaMalloc(&xd, n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&yd, n*sizeof(float)) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(xd, x, n*sizeof(float), cudaMemcpyHostToDevice) );

	blocksize = (n+nblocks-1)/nblocks;
    for (i=0; i<niters; i++) {
        cuda_saxpb<<<nblocks, blocksize>>>(xd, a, b, yd, n);
    }
	CHK_CUDA( cudaMemcpy(ycuda, yd, n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	
	CHK_CUDA( cudaFree(xd) );
	CHK_CUDA( cudaFree(yd) );

    abserr = 0.;
    for (i=0;i<n;i++) {
        abserr += fabs(ycuda[i] - y[i]);
    }
    
    printf("CPU time = %lg millisec.\n", cputime*1000.);
    printf("GPU time = %lg millisec.\n", gputime*1000.);
    printf("CUDA and CPU results differ by %lf\n", abserr);

    free(x);
    free(y);
    free(ycuda);
	return 0;
}


int get_options(int argc, char **argv, int *n, int *nb, float *a, float *b, int *niters) {

  const struct option long_options[] = {
    {"nvals"     , required_argument, 0, 'n'},
    {"nblocks"   , required_argument, 0, 'B'},
    {"a", required_argument, 0, 'a'},
    {"b", required_argument, 0, 'b'},
    {"niters", required_argument, 0, 'I'},
    {"help",      no_argument, 0, 'h'},
    {0, 0, 0, 0}};

  char c;
  int option_index;
  int tempint;

  while (1) {
    c = getopt_long(argc, argv, "n:B:a:b:h", long_options, &option_index);
    if (c == -1) break;
 
    switch(c) {
      case 'n': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 500000) {
            fprintf(stderr,"%s: Cannot use number of points %s;\n  Using %d\n", argv[0], optarg, *n);
          } else {
            *n = tempint;
          }
          break;

      case 'B': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 1000 || tempint > *n) {
            fprintf(stderr,"%s: Cannot use number of blocks %s;\n  Using %d\n", argv[0], optarg, *nb);
          } else {
            *nb = tempint;
          }
          break;

      case 'I': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 1000 || tempint > *n) {
            fprintf(stderr,"%s: Cannot use number of iterations %s;\n  Using %d\n", argv[0], optarg, *niters);
          } else {
            *niters = tempint;
          }
          break;

      case 'a': *a = atof(optarg);
          break;

      case 'b': *b = atof(optarg);
          break;

      case 'h':
          puts("Calculates y[i] = a*x[i] + b on the GPU.");
          puts("Options: ");
          puts("    --nvals=N      (-n N): Set the number of values in y,x.");
          puts("    --nblocks=N    (-B N): Set the number of blocks used.");
          puts("    --a=X          (-a X): Set the parameter a.");
          puts("    --b=X          (-b X): Set the parameter b.");
          puts("    --niters=N     (-I X): Set number of iterations to calculate.");
          puts("");
          return +1;
        }
    }

    return 0;
}

void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}
