#include <stdio.h>
#include <stdlib.h>
#include <math.h>

unsigned char charmean(unsigned char *data, const int r, const int c, const int nhalo, const int rowsize) {

    const float sig = 2./9.*(nhalo*nhalo);
    float pi = 4.*atan(1.);

    float mean=0.;
    for (int i=r-nhalo; i<=r+nhalo; i++) {
        for (int j=c-nhalo; j<=c+nhalo; j++) {
            float d  = 1.0*data[i*rowsize+j];
            float e = expf(-(float)((i-r)*(i-r)+(j-c)*(j-c))/sig);
            mean += d*e;
        }
    }

    mean /= sig*pi;

    return (unsigned char)roundf(mean);
}


void smoothimage(unsigned char *restrict data, unsigned char *restrict smootheddata, const int nhalos, const int rows, const int cols) {

    const int rowsize = rows+2*nhalos;

    for (int i=nhalos; i<nhalos+rows; i++) {
        for (int j=nhalos; j<nhalos+cols; j++) {
            smootheddata[i*rowsize+j] = charmean(data, i, j, nhalos, rowsize);
        }
    }
}


size_t pgmWriteWithHalo(char *fname, const int cols, const int rows, 
                           const int max, const int nhalo, unsigned char **data);

size_t pgmReadWithHalo(char *fname, int *cols, int *rows, 
                           int *max, int nhalo, unsigned char ***data);

unsigned char **char2darray(int n, int m);

void free2darray(unsigned char **array);

void setchar2darray(unsigned char **array, int n, int m, unsigned char val);


int main(int argc, char **argv) {
    
    if (argc != 4) {
        fprintf(stderr,"Usage: %s infile.pgm outfile.pgm filtersize\n", argv[0]);
        fprintf(stderr,"       Smooths infile with median filter of size filtersize, writes to outfile.\n");
        exit(-1);
    }

    unsigned char **data;
    int  cols, rows, max;

    int nhalos = atoi(argv[3]);
    const int defnhalos = 3;
    const int maxnhalos = 50;
    if (nhalos < 0 || nhalos > maxnhalos) {
        fprintf(stderr,"%s: filter size %s invalid; using default %d\n", argv[0], argv[3], defnhalos);
        nhalos = defnhalos;
    }

    int nread = pgmReadWithHalo(argv[1], &cols, &rows, &max, nhalos, &data);
    if (nread == 0) {
        fprintf(stderr,"%s: Error reading file %s\n", argv[0], argv[1]);
        exit(-2);
    }

    unsigned char **smootheddata = char2darray(rows+2*nhalos, cols+2*nhalos);
    smoothimage(&(data[0][0]), &(smootheddata[0][0]), nhalos, rows, cols);

    int nwrote = pgmWriteWithHalo(argv[2], cols, rows, max, nhalos, smootheddata);

    if (nwrote == 0) {
        fprintf(stderr,"%s: Error writing file %s\n", argv[0], argv[2]);
        free2darray(smootheddata); 
        free2darray(data); 
        exit(-3);
    }

    free2darray(smootheddata); 
    free2darray(data); 
    return 0;
}


/*
 * Utility routines used by the main program for allocating 2d unsigned character
 * arrays, and reading/writing image files in PGM format.
 */

#define PGMMAGIC "P5"

void getNonCommentLine(FILE *in, char *line, const int len) {
    do {
        fgets(line, len, in);
    } while (line[0] == '#');
}

void scanHeader(FILE *in, char *magic, int *col, int *row, int *max) {
    const int len=128;
    char dummy[len+1];

    getNonCommentLine(in, dummy, len);
    sscanf(dummy, "%s\n", magic);

    getNonCommentLine(in, dummy, len);
    sscanf(dummy, "%d %d\n", col, row);
    
    getNonCommentLine(in, dummy, len);
    sscanf(dummy, "%d\n", max);
}

int pgmReadHeader(char *fname, char *magic, int *col, int *row, int *max) {
    
    FILE *in = fopen(fname,"r");
    if (!in) 
        return -1;

    scanHeader(in, magic, col, row, max);
    fclose(in);

    return 0;
}

void skipheader(FILE *in) {
    char magic[3];
    int col, row, max; 
    scanHeader(in, magic, &col, &row, &max);
}

#include <string.h>

int correctmagic(char *magic) {
    return !strncmp(magic, PGMMAGIC, 2);
}

size_t pgmReadData (char *fname, const int cols, const int rows, unsigned char *data) {
    size_t nread;
    
    FILE *in = fopen(fname,"r");
    if (!in) 
        return -1;

    skipheader(in);

    nread = fread(data, sizeof(unsigned char), cols*rows, in);

    fclose(in);
    return nread;
}


size_t pgmReadDataWithHalo(char *fname, const int cols, const int rows, 
                           int nhalo, unsigned char **data) {
    size_t nread = 0;
    
    FILE *in = fopen(fname,"r");
    if (!in) 
        return -1;

    skipheader(in);

    for (int row=0; row<rows; row++) 
        nread += fread(&(data[row+nhalo][nhalo]), sizeof(unsigned char), cols, in);

    fclose(in);
    return nread;
}

void writeHeader(FILE *out, int cols, int rows, int max) {

    fprintf(out, "%s\n", PGMMAGIC);
    fprintf(out, "%d %d\n", cols, rows);
    fprintf(out, "%d\n", max);

}

size_t pgmWrite(char *fname, const int cols, const int rows, 
                     const int max, unsigned char *data) {
    size_t nwrote;
    
    FILE *out = fopen(fname,"wb");
    if (!out) 
        return -1;

    writeHeader(out, cols, rows, max);
    nwrote = fwrite(data, sizeof(unsigned char), cols*rows, out);

    fclose(out);
    return nwrote;
}


size_t pgmWriteWithHalo(char *fname, const int cols, const int rows, 
                           const int max, const int nhalo, unsigned char **data) {
    FILE *out = fopen(fname,"wb");
    
    if (!out) 
        return -1;

    writeHeader(out, cols, rows, max);

    size_t nwrote = 0;
    for (int row=0; row<rows; row++) 
        nwrote += fwrite(&(data[row+nhalo][nhalo]), sizeof(unsigned char), cols, out);

    fclose(out);
    return nwrote;
}

size_t pgmRead(char *fname, int *cols, int *rows, int *max, unsigned char ***data) {

    char magic[6];
    int res = pgmReadHeader(fname, magic, cols, rows, max);
    if (res < 0)
        return 0;

    if (!correctmagic(magic))
        return 0;

    *data = char2darray(*rows, *cols);
    return pgmReadData (fname, *cols, *rows, &((*data)[0][0]));
}


size_t pgmReadWithHalo(char *fname, int *cols, int *rows, 
                           int *max, int nhalo, unsigned char ***data) {

    char magic[6];
    int res = pgmReadHeader(fname, magic, cols, rows, max);
    if (res < 0)
        return 0;

    if (!correctmagic(magic))
        return 0;

    *data = char2darray((*rows)+2*nhalo, (*cols)+2*nhalo);
    setchar2darray(*data, (*rows)+2*nhalo, (*cols)+2*nhalo, 0);

    return pgmReadDataWithHalo (fname, *cols, *rows, nhalo, *data);
}

unsigned char **char2darray(int n, int m) {
    unsigned char *data = malloc(n*m*sizeof(unsigned char));
    unsigned char **array = malloc(n*sizeof(unsigned char*));
    for (int i=0; i<n; i++)
        array[i] = &(data[i*m]);

    return array;
}

void free2darray(unsigned char **array) {
    free(array[0]);
    free(array);
}

void setchar2darray(unsigned char **array, int n, int m, unsigned char val) {
    memset(array[0], val, n*m);
}
